"""
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
See the GNU General Public License for more details.


You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.

Copyright © 2019 Cloud Linux Software Inc.

This software is also available under ImunifyAV commercial license,
see <https://www.imunify360.com/legal/eula>
"""
import asyncio
import time
from contextlib import suppress
from pathlib import Path

from defence360agent.api.server.cleanup_revert import CleanupRevertAPI
from defence360agent.contracts.config import Core
from defence360agent.contracts.license import LicenseCLN
from defence360agent.contracts.plugins import MessageSource
from imav.malwarelib.cleanup.storage import restore_hits
from imav.malwarelib.config import MalwareHitStatus
from imav.malwarelib.model import MalwareHit
from defence360agent.utils import recurring_check

PERIOD = 3 * 60 * 60  # 3 hours


class CleanupRevertPlugin(MessageSource):
    initiator = "imunify"
    _sink = _task = None

    def __init__(self):
        self._cleanup_revert_flag = Path(Core.TMPDIR) / "cleanup_revert"

    async def create_source(self, loop, sink):
        self._sink = sink
        self._task = loop.create_task(
            recurring_check(PERIOD)(self.process_cleanup_revert)()
        )

    async def shutdown(self):
        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def process_cleanup_revert(self):
        if (
            self._cleanup_revert_flag.exists()
            and (self._cleanup_revert_flag.stat().st_mtime + PERIOD)
            > time.time()
        ):
            return

        if LicenseCLN.is_free():
            return

        if files := await CleanupRevertAPI.paths():
            if hits := list(
                MalwareHit.get_hits(
                    files, statuses=MalwareHitStatus.RESTORABLE
                )
            ):
                await restore_hits(hits, self._sink, self.initiator)

        self._cleanup_revert_flag.touch(mode=0o644, exist_ok=True)
